import OUtils.configurations
import OUtils.Kafka.metrics as metrics
import OUtils.Kafka.checks as checks


def get_nodes(port=False):
    result = []
    nodes = OUtils.configurations.get(module="kafka", key="nodes")

    for node in nodes:
        if port is False:
            result.append(node["host"])

        elif port not in node["ports"]:
            raise KeyError("Port '%s' not supported" % port)

        else:
            result.append("%s:%d" % (node["host"], node["ports"][port]))

    assert len(result) >= 3, "Too few nodes (%s) loaded. Please check configuration file." % len(result)

    return result


def get_configuration_files():
    result = OUtils.configurations.get(module="kafka", key="configuration_files")

    assert len(result) > 0, "Too few (%d) configuration files loaded. Please check configurations file." % len(result)

    return result



import OUtils.Kafka

import subprocess
import time


def __verify_cluster_clean_state__():
    assert OUtils.Kafka.checks.are_configuration_files_aligned() is True, "Configuration files not aligned"
    assert OUtils.Kafka.metrics.get_total_under_replicated_replicas() == 0, "There are under-replicated replicas"

    return True


def __get_confirmation__(string):
    confirmation = input(string)

    if confirmation.lower() in ["yes", "y"]:
        return True

    return False


def __execute_kafka_service_restart__(node):
    subprocess.run("bash -c \"ssh root@%s 'systemctl restart kafka'\"" % node, shell=True, check=True)

    return True


def roll_restart():
    try:
        # Step 1
        if __get_confirmation__("Please confirm rolling restart of '%s': " % OUtils.Kafka.get_nodes()) is False:
            print("Rolling restart aborted")
            return

        # Step 2
        print("Verifying clean cluster status before restart ...")

        if __verify_cluster_clean_state__() is True:
            print("\tOK")

        # Step 3
        print("Executing rolling restart ...")

        for node in OUtils.Kafka.get_nodes():

            # Step 3a: restart service
            print("\tOn node %s: Restarting Kafka service ..." % node)
            if __execute_kafka_service_restart__(node) is True:
                print("\t\tOK")

            # Step 3b: check that cluster status returns to clean
            print("\tOn node %s: Checking under-replication ..." % node)

            for attempt in range(10):
                time.sleep(5)

                try:
                    under_replicated_replicas = OUtils.Kafka.metrics.get_total_under_replicated_replicas()

                except ConnectionError:
                    print("\t\tFailed to connect. Will re-try.")
                    continue

                if under_replicated_replicas == 0:
                    print("\t\tOK: No under-replicated replicas found: system stabilized back.\n")
                    break

                print("\t\tFound %d under-replicated partitions." % under_replicated_replicas)
                continue

            if under_replicated_replicas != 0:
                raise Exception("System did not stabilize back after several attempts")

        print("Rolling restart completed.")
        return True

    except Exception as err:
        print("An exception occurred and the rolling restart was interrupted. Please verify cluster status.")
        print(str(err))
        return False



import OUtils.Kafka

import subprocess
import re


def __is_file_aligned_across_nodes__(file_path, whitelisted_patterns):
    assert file_path.startswith("/"), "The value of filePath (%s) appears invalid. Please check." % file_path

    nodes = OUtils.Kafka.get_nodes()

    # Compare each file against the copy on another's node (there is 1x overlap!)
    for i in range(len(nodes)):
        node_a = nodes[i - 1]
        node_b = nodes[i]

        cmd = "bash -c \"diff <(ssh %s 'cat %s') <(ssh %s 'cat %s') | grep -E '< .|> .' | cut -c 3-\""
        cmd = cmd % (node_a, file_path, node_b, file_path)

        output = subprocess.run(cmd, shell=True, capture_output=True)

        if len(output.stderr.decode("utf-8")) > 0:
            raise Exception("Got error '%s' while executing '%s'" % (output.stderr.decode("utf-8"), cmd))

        differences = output.stdout.decode("utf-8").splitlines()

        # Each difference found must be accepted by any of the whitelisting rules provided for this file
        for difference in differences:
            if __validate_difference__(difference, whitelisted_patterns) is True:
                continue

            raise Exception("Difference '%s' found on [%s, %s] for file '%s'" % (difference, node_a, node_b, file_path))

    return True


def __validate_difference__(difference, whitelisted_patterns):
    for pattern in whitelisted_patterns:
        if re.fullmatch(pattern, difference) is not None:
            return True

    return False


def are_configuration_files_aligned():
    files_to_check = OUtils.Kafka.get_configuration_files()

    for file in files_to_check:
        file_path = file["file_path"]
        whitelisted_patterns = file["whitelisted_patterns"]

        if __is_file_aligned_across_nodes__(file_path=file_path, whitelisted_patterns=whitelisted_patterns) is False:
            return False

    return True



import os
import json
import jsonschema

__configurations_file_path__ = './configurations.json'
__loaded_configurations__ = None


def __load_configurations__():
    global __loaded_configurations__

    if __loaded_configurations__ is not None:
        return

    if not os.path.exists(__configurations_file_path__):
        raise FileExistsError("Configuration file '%s' not found" % __configurations_file_path__)

    with open(__configurations_file_path__) as json_file:
        __loaded_configurations__ = json.load(json_file)


def __validate_configurations__():
    global __loaded_configurations__

    if __loaded_configurations__ is None:
        raise ImportError("Can't validate configurations (Not loaded yet")

    schema_file_path = os.path.dirname(os.path.realpath(__file__)) + '/configurationSchema.json'

    with open(schema_file_path) as json_file:
        schema = json.load(json_file)

        jsonschema.validate(instance=__loaded_configurations__, schema=schema)


def get(module, key):
    global __loaded_configurations__

    if __loaded_configurations__ is None:
        __load_configurations__()
        __validate_configurations__()

    if module not in __loaded_configurations__:
        raise KeyError("Configuration module '%s' not found" % module)

    if key not in __loaded_configurations__[module]:
        raise KeyError("Configuration '%s' not found under module '%s'" % (key, module))

    return __loaded_configurations__[module][key]



import setuptools

setuptools.setup(
	name="OUtils",
	version="0.0.15",
	author="Schups",
	author_email="schups@gmail.com",
	description="Operation Utilities - Python3 package containing common scripts for managing or monitoring various services.",
	packages=setuptools.find_packages(),
	python_requires='>=3.7',
	include_package_data=True,
        install_requires=['jsonschema', 'jmxquery']
)



import OUtils.Kafka

import subprocess
import jmxquery
import sys
import os


def __execute_jmx_query__(node, query_string):
    assert ":" in node, "Missing port information on provided argument '%s'" % node

    try:
        # This is suppress visualization of jar stderr errors (e.g. connection failed)
        # Ref. https://stackoverflow.com/questions/2125702/how-to-suppress-console-output-in-python
        # Ref. https://stackoverflow.com/questions/1956142/how-to-redirect-stderr-in-python
        sys.stderr = open(os.devnull, "w")

        jmx_connection = jmxquery.JMXConnection(connection_uri="service:jmx:rmi:///jndi/rmi://%s/jmxrmi" % node)
        jmx_results = jmx_connection.query([jmxquery.JMXQuery(mBeanName=query_string)])

    except subprocess.SubprocessError as error:
        if "ConnectException" in error.output.decode("utf-8"):
            raise ConnectionError(error.output.decode("utf-8"))

        else:
            raise error

    finally:
        # Conclusion of the trick above
        sys.stderr = sys.__stderr__

    return jmx_results


def get_under_replicated_replicas_per_node(node):
    results = __execute_jmx_query__(node, "kafka.server:type=ReplicaManager,name=UnderReplicatedPartitions")

    if len(results) == 0:
        raise ConnectionError("Unable to load JMX Query results (timeout?)")

    return results[0].value


def get_total_under_replicated_replicas():
    result = 0

    for node in OUtils.Kafka.get_nodes(port="jmx"):
        result = get_under_replicated_replicas_per_node(node=node)

    return result
